
import datetime
from typing import Optional
from pydantic import BaseModel, Field

class TraceModel(BaseModel):
    name: str
    start_time: str = Field(default_factory=datetime.datetime.utcnow)
    end_time: Optional[str] = None
    duration: Optional[int] = None
    status: Optional[str] = None
    attributes: Optional[dict] = None
    input: Optional[dict] = None
    output: Optional[dict] = None
    version: Optional[str] = None

class SpanModel(BaseModel):
    name: str
    start_time: str = Field(default_factory=datetime.datetime.utcnow)
    span_type: str = "span"
    end_time: Optional[str] = None
    duration: Optional[int] = None
    status: Optional[str] = None
    attributes: Optional[dict] = None
    input: Optional[dict] = None
    output: Optional[dict] = None
    version: Optional[str] = None



import datetime
from typing import Any, Dict, List, Optional, Union
from .models import SpanModel
from .util import get_utc_time, remove_none_values
from langchain.schema.document import Document

class Span:

    def __init__(
        self,
        name: str,
        start_time: Optional[datetime.datetime] = None,
        end_time: Optional[datetime.datetime] = None,
        span_type: str = "span",
        status: Optional[str] = None,
        attributes: Optional[dict] = None,
        input: Optional[dict] = None,
        output: Optional[dict] = None,
        duration: Optional[int] = None,
        version: Optional[str] = None,
    ):
        if start_time is None:
            start_time = datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc)
        else:
            start_time = start_time
        self._span = SpanModel(
            name=name,
            span_type=span_type,
            start_time=start_time.isoformat(),
            end_time=end_time.isoformat() if end_time else None,
            status=status,
            attributes=attributes or {},
            input=input or {},
            output=output or {},
            duration=duration,
            version=version,
        )
        self._children = []

    def __repr__(self):
        return f"Span(name={self._span.name}, dict={remove_none_values(self.to_dict())}, children={self._children})"

    def to_dict(self):
        span_dict = self._span.model_dump()
        if "input_documents" in self._span.input:
            self._span.input['input_documents'] = [doc.page_content if isinstance(doc, Document) else doc for doc in self._span.input["input_documents"]]
        span_dict["children"] = [child.to_dict() for child in self._children]
        return span_dict

    def add_span(self, span: "Span"):
        self._children.append(span)

    def create_span(
        self,
        name: str,
        start_time: Optional[datetime.datetime] = None,
        end_time: Optional[datetime.datetime] = None,
        span_type: str = "span",
        status: Optional[str] = None,
        attributes: Optional[dict] = None,
        input: Optional[dict] = None,
        output: Optional[dict] = None,
        duration: Optional[int] = None,
        version: Optional[str] = None,
    ):
        span = Span(
            name=name,
            start_time=(start_time or datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc)),
            end_time=end_time,
            span_type=span_type,
            status=status,
            attributes=attributes or {},
            input=input or {},
            output=output or {},
            duration=duration,
            version=version,
        )
        self._children.append(span)
        return span

    def add_generation(self, generation: "Generation"):
        self._children.append(generation)

    def create_generation(
        self,
        name: str,
        start_time: Optional[datetime.datetime] = None,
        end_time: Optional[datetime.datetime] = None,
        span_type: str = "generation",
        status: Optional[str] = None,
        attributes: Optional[dict] = None,
        input: Optional[dict] = None,
        output: Optional[dict] = None,
        duration: Optional[int] = None,
        version: Optional[str] = None,
        prompt: Optional[Union[List[Dict[str, Any]], Dict[str, Any], str]] = None,
        response: Optional[Any] = None,
        prompt_slug: Optional[str] = None,
        language_model_id: Optional[str] = None,
        environment: Optional[str] = None,
        functions: Optional[List[Dict]] = None,
        function_call_response: Optional[Any] = None,
        tools: Optional[Any] = None,
        tool_calls: Optional[Any] = None,
        external_reference_id: Optional[str] = None,
        customer_id: Optional[str] = None,
        customer_user_id: Optional[str] = None,
        session_id: Optional[str] = None,
        user_query: Optional[str] = None,
        prompt_tokens: Optional[int] = None,
        completion_tokens: Optional[int] = None,
        total_tokens: Optional[int] = None,
        response_time: Optional[int] = None,
        context: Optional[Dict] = None,
        expected_response: Optional[str] = None,
        custom_attributes: Optional[Dict] = None,
        cost: Optional[float] = None,
        custom_eval_metrics: Optional[Dict] = None,
    ):
        span = Generation(
            name=name,
            start_time=start_time,
            end_time=end_time,
            span_type=span_type,
            status=status,
            attributes=attributes,
            input=input,
            output=output,
            duration=duration,
            version=version,
            prompt=prompt,
            response=response,
            prompt_slug=prompt_slug,
            language_model_id=language_model_id,
            environment=environment,
            functions=functions,
            function_call_response=function_call_response,
            tools=tools,
            tool_calls=tool_calls,
            external_reference_id=external_reference_id,
            customer_id=customer_id,
            customer_user_id=customer_user_id,
            session_id=session_id,
            user_query=user_query,
            prompt_tokens=prompt_tokens,
            completion_tokens=completion_tokens,
            total_tokens=total_tokens,
            response_time=response_time,
            context=context,
            expected_response=expected_response,
            custom_attributes=custom_attributes,
            cost=cost,
            custom_eval_metrics=custom_eval_metrics,
        )
        self._children.append(span)
        return span
    
    def update(
        self,
        end_time: Optional[datetime.datetime] = None,
        status: Optional[str] = None,
        input: Optional[Dict[str, Any]] = None,
        output: Optional[Dict[str, Any]] = None,
        duration: Optional[int] = None,
        attributes: Optional[Dict[str, Any]] = None,
    ):
        if end_time:
            self._span.end_time = end_time.replace(tzinfo=datetime.timezone.utc)
        if status:
            self._span.status = status
        if input:
            self._span.input = input
        if output:
            self._span.output = output
        if duration:
            self._span.duration = duration
        if attributes:
            self._span.attributes.update(attributes)

    def end(self, end_time: Optional[datetime.datetime] = None):
        try:
            end_time = get_utc_time(end_time)
            if self._span.end_time is None:
                self._span.end_time = end_time.replace(tzinfo=datetime.timezone.utc).isoformat()
            if self._span.duration is None:
                delta = (end_time - get_utc_time(datetime.datetime.fromisoformat(self._span.start_time)))
                self._span.duration = int((delta.seconds * 1000) + (delta.microseconds // 1000))
            for child in self._children:
                child.end(end_time)
        except Exception as e:
            print(f"Error ending span: {e}")

class Generation(Span):
    def __init__(
        self,
        name: str,
        start_time: Optional[datetime.datetime] = None,
        end_time: Optional[datetime.datetime] = None,
        span_type: str = "generation",
        status: Optional[str] = None,
        attributes: Optional[Dict[str, Any]] = None,
        input: Optional[Dict[str, Any]] = None,
        output: Optional[Dict[str, Any]] = None,
        duration: Optional[int] = None,
        version: Optional[str] = None,
        prompt: Optional[Union[List[Dict[str, Any]], Dict[str, Any], str]] = None,
        response: Optional[Any] = None,
        prompt_slug: Optional[str] = None,
        language_model_id: Optional[str] = None,
        environment: Optional[str] = None,
        functions: Optional[List[Dict]] = None,
        function_call_response: Optional[Any] = None,
        tools: Optional[Any] = None,
        tool_calls: Optional[Any] = None,
        external_reference_id: Optional[str] = None,
        customer_id: Optional[str] = None,
        customer_user_id: Optional[str] = None,
        session_id: Optional[str] = None,
        user_query: Optional[str] = None,
        prompt_tokens: Optional[int] = None,
        completion_tokens: Optional[int] = None,
        total_tokens: Optional[int] = None,
        response_time: Optional[int] = None,
        context: Optional[Dict] = None,
        expected_response: Optional[str] = None,
        custom_attributes: Optional[Dict] = None,
        cost: Optional[float] = None,
        custom_eval_metrics: Optional[Dict] = None,
    ):
        if attributes is None:
            attributes = {}
        additional_attributes = {
            "prompt": prompt,
            "response": response,
            "prompt_slug": prompt_slug,
            "language_model_id": language_model_id,
            "environment": environment,
            "functions": functions,
            "function_call_response": function_call_response,
            "tools": tools,
            "tool_calls": tool_calls,
            "external_reference_id": external_reference_id,
            "customer_id": customer_id,
            "customer_user_id": customer_user_id,
            "session_id": session_id,
            "user_query": user_query,
            "prompt_tokens": prompt_tokens,
            "completion_tokens": completion_tokens,
            "total_tokens": total_tokens,
            "response_time": response_time,
            "context": context,
            "expected_response": expected_response,
            "custom_attributes": custom_attributes,
            "cost": cost,
            "custom_eval_metrics": custom_eval_metrics,
        }
        # Update 'attributes' only with non-None values from 'additional_attributes'
        for key, value in additional_attributes.items():
            if value is not None:
                attributes[key] = value
        super().__init__(
            name=name,
            start_time=start_time,
            end_time=end_time,
            span_type=span_type,
            status=status,
            attributes=attributes,
            input=input,
            output=output,
            duration=duration,
            version=version,
        )

    def update(
        self,
        end_time: Optional[datetime.datetime] = None,
        status: Optional[str] = None,
        input: Optional[Dict[str, Any]] = None,
        output: Optional[Dict[str, Any]] = None,
        duration: Optional[int] = None,
        prompt: Optional[Union[List[Dict[str, Any]], Dict[str, Any], str]] = None,
        response: Optional[Any] = None,
        prompt_slug: Optional[str] = None,
        language_model_id: Optional[str] = None,
        environment: Optional[str] = None,
        functions: Optional[List[Dict]] = None,
        function_call_response: Optional[Any] = None,
        tools: Optional[Any] = None,
        tool_calls: Optional[Any] = None,
        external_reference_id: Optional[str] = None,
        customer_id: Optional[str] = None,
        customer_user_id: Optional[str] = None,
        session_id: Optional[str] = None,
        user_query: Optional[str] = None,
        prompt_tokens: Optional[int] = None,
        completion_tokens: Optional[int] = None,
        total_tokens: Optional[int] = None,
        response_time: Optional[int] = None,
        context: Optional[Dict] = None,
        expected_response: Optional[str] = None,
        custom_attributes: Optional[Dict] = None,
        cost: Optional[float] = None,
        custom_eval_metrics: Optional[Dict] = None,
    ):
        if self._span.attributes is None:
            self._span.attributes = {}
        attributes = {
            "prompt": prompt if prompt is not None else self._span.attributes.get("prompt"),
            "response": response if response is not None else self._span.attributes.get("response"),
            "prompt_slug": prompt_slug if prompt_slug is not None else self._span.attributes.get("prompt_slug"),
            "language_model_id": language_model_id if language_model_id is not None else self._span.attributes.get("language_model_id"),
            "environment": environment if environment is not None else self._span.attributes.get("environment"),
            "functions": functions if functions is not None else self._span.attributes.get("functions"),
            "function_call_response": function_call_response if function_call_response is not None else self._span.attributes.get("function_call_response"),
            "tools": tools if tools is not None else self._span.attributes.get("tools"),
            "tool_calls": tool_calls if tool_calls is not None else self._span.attributes.get("tool_calls"),
            "external_reference_id": external_reference_id if external_reference_id is not None else self._span.attributes.get("external_reference_id"),
            "customer_id": customer_id if customer_id is not None else self._span.attributes.get("customer_id"),
            "customer_user_id": customer_user_id if customer_user_id is not None else self._span.attributes.get("customer_user_id"),
            "session_id": session_id if session_id is not None else self._span.attributes.get("session_id"),
            "user_query": user_query if user_query is not None else self._span.attributes.get("user_query"),
            "prompt_tokens": prompt_tokens if prompt_tokens is not None else self._span.attributes.get("prompt_tokens"),
            "completion_tokens": completion_tokens if completion_tokens is not None else self._span.attributes.get("completion_tokens"),
            "total_tokens": total_tokens if total_tokens is not None else self._span.attributes.get("total_tokens"),
            "response_time": response_time if response_time is not None else self._span.attributes.get("response_time"),
            "context": context if context is not None else self._span.attributes.get("context"),
            "expected_response": expected_response if expected_response is not None else self._span.attributes.get("expected_response"),
            "custom_attributes": custom_attributes if custom_attributes is not None else self._span.attributes.get("custom_attributes"),
            "cost": cost if cost is not None else self._span.attributes.get("cost"),
            "custom_eval_metrics": custom_eval_metrics if custom_eval_metrics is not None else self._span.attributes.get("custom_eval_metrics"),
        }
        super().update(
            end_time=end_time,
            status=status,
            input=input,
            output=output,
            duration=duration,
            attributes=attributes,
        )



import asyncio
import json
import datetime
import threading
from typing import Any, Dict, List, Optional, Union

from .span import Generation, Span
from .models import TraceModel
from .util import get_utc_time, remove_none_values, sanitize_dict
from athina_logger.api_key import AthinaApiKey
from athina_logger.constants import API_BASE_URL
from athina_logger.request_helper import RequestHelper
from langchain.schema.document import Document

class Trace(AthinaApiKey):
    _trace: TraceModel
    _spans: list[Span]

    def __init__(
        self,
        name: str,
        start_time: Optional[datetime.datetime] = None,
        end_time: Optional[datetime.datetime] = None,
        status: Optional[str] = None,
        attributes: Optional[dict] = None,
        input: Optional[dict] = None,
        output: Optional[dict] = None,
        duration: Optional[int] = None,
        version: Optional[str] = None,
    ):        
        self._trace = TraceModel(
            name=name,
            start_time=(start_time or datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc)).isoformat(),
            end_time=end_time.isoformat() if end_time else None,
            status=status,
            attributes=attributes or {},
            input=input or {},
            output=output or {},
            duration=duration,
            version=version,
        )
        self._spans = []

    def __repr__(self):
        return f"Trace(name={self._trace.name}, dict={remove_none_values(self.to_dict())},  spans={self._spans})"

    def add_span(self, span: Span):
        self._spans.append(span)

    def create_span(
        self,
        name: str,
        start_time: Optional[datetime.datetime] = None,
        end_time: Optional[datetime.datetime] = None,
        span_type: str = "span",
        status: Optional[str] = None,
        attributes: Optional[dict] = None,
        input: Optional[dict] = None,
        output: Optional[dict] = None,
        duration: Optional[int] = None,
        version: Optional[str] = None,
    )-> Span:
        span = Span(
            name=name,
            start_time=(start_time or datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc)),
            end_time=end_time,
            span_type=span_type,
            status=status,
            attributes=attributes or {},
            input=input or {},
            output=output or {},
            duration=duration,
            version=version,
        )
        self._spans.append(span)
        return span

    def add_generation(self, generation: Generation):
        self._spans.append(generation)

    def create_generation(
        self,
        name: str,
        start_time: Optional[datetime.datetime] = None,
        end_time: Optional[datetime.datetime] = None,
        span_type: str = "generation",
        status: Optional[str] = None,
        attributes: Optional[dict] = None,
        input: Optional[dict] = None,
        output: Optional[dict] = None,
        duration: Optional[int] = None,
        version: Optional[str] = None,
        prompt: Optional[Union[List[Dict[str, Any]], Dict[str, Any], str]] = None,
        response: Optional[Any] = None,
        prompt_slug: Optional[str] = None,
        language_model_id: Optional[str] = None,
        environment: Optional[str] = None,
        functions: Optional[List[Dict]] = None,
        function_call_response: Optional[Any] = None,
        tools: Optional[Any] = None,
        tool_calls: Optional[Any] = None,
        external_reference_id: Optional[str] = None,
        customer_id: Optional[str] = None,
        customer_user_id: Optional[str] = None,
        session_id: Optional[str] = None,
        user_query: Optional[str] = None,
        prompt_tokens: Optional[int] = None,
        completion_tokens: Optional[int] = None,
        total_tokens: Optional[int] = None,
        response_time: Optional[int] = None,
        context: Optional[Dict] = None,
        expected_response: Optional[str] = None,
        custom_attributes: Optional[Dict] = None,
        cost: Optional[float] = None,
        custom_eval_metrics: Optional[Dict] = None,
    ) -> Generation:
        span = Generation(
            name=name,
            start_time=(start_time or datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc)),
            end_time=end_time,
            span_type=span_type,
            status=status,
            attributes=attributes,
            input=input,
            output=output,
            duration=duration,
            version=version,
            prompt=prompt,
            response=response,
            prompt_slug=prompt_slug,
            language_model_id=language_model_id,
            environment=environment,
            functions=functions,
            function_call_response=function_call_response,
            tools=tools,
            tool_calls=tool_calls,
            external_reference_id=external_reference_id,
            customer_id=customer_id,
            customer_user_id=customer_user_id,
            session_id=session_id,
            user_query=user_query,
            prompt_tokens=prompt_tokens,
            completion_tokens=completion_tokens,
            total_tokens=total_tokens,
            response_time=response_time,
            context=context,
            expected_response=expected_response,
            custom_attributes=custom_attributes,
            cost=cost,
            custom_eval_metrics=custom_eval_metrics,
        )
        self._spans.append(span)
        return span
    
    def to_dict(self):
        trace_dict = self._trace.model_dump()
        trace_dict["spans"] = [span.to_dict() for span in self._spans]
        return trace_dict

    def update(
        self,
        end_time: Optional[datetime.datetime] = None,
        duration: Optional[int] = None,
        status: Optional[str] = None,
        attributes: Optional[Dict[str, Any]] = None,
    ):
        if end_time:
            self._trace.end_time = end_time.replace(tzinfo=datetime.timezone.utc).isoformat()
        if status:
            self._trace.status = status
        if attributes:
            self._trace.attributes = attributes
        if duration:
            self._trace.duration = duration

    async def _log_trace_async(self, request_dict: Dict[str, Any]):
        RequestHelper.make_post_request(endpoint=f'{API_BASE_URL}/api/v1/trace/sdk', payload=request_dict, headers={
            "athina-api-key": Trace.get_api_key(), "Content-Type": "application/json"
        })

    def end(self, end_time: Optional[datetime.datetime] = None):
        try:      
            end_time = get_utc_time(end_time)
            for span in self._spans:
                span.end(end_time)
            if self._trace.end_time is None:
                self._trace.end_time =  end_time.replace(tzinfo=datetime.timezone.utc).isoformat()
            if self._trace.duration is None:
                delta = (end_time - get_utc_time(datetime.datetime.fromisoformat(self._trace.start_time)))
                self._trace.duration = int((delta.seconds * 1000) + (delta.microseconds // 1000))
            request_dict = remove_none_values(self.to_dict())
            request_dict = sanitize_dict(request_dict)
            threading.Thread(target=lambda: asyncio.run(self._log_trace_async(request_dict))).start()
        except Exception as e:
            print("Error ending trace: ", e)


from datetime import datetime, timezone
import json

def remove_none_values(d):
    if isinstance(d, dict):
        return {k: remove_none_values(v) for k, v in d.items() if v is not None}
    elif isinstance(d, list):
        return [remove_none_values(v) for v in d if v is not None]
    else:
        return d

def sanitize_dict(d):
    """Recursively sanitize dictionary to ensure JSON serializability."""
    if isinstance(d, dict):
        return {k: sanitize_dict(v) for k, v in d.items()}
    elif isinstance(d, list):
        return [sanitize_dict(item) for item in d]
    elif hasattr(d, '__dict__'):  # Custom object
        return f"<{d.__class__.__name__} object>"
    else:
        # Handle basic types that are JSON serializable
        try:
            json.dumps(d)
            return d
        except (TypeError, OverflowError):
            return str(d)  # Convert to string as fallback

def get_utc_time(time_obj=None):
    if time_obj is None:
        return datetime.now(timezone.utc)
    if time_obj.tzinfo is None:
        time_obj = time_obj.replace(tzinfo=timezone.utc)
    return time_obj.astimezone(timezone.utc)



import re
from typing import Any, Dict, List, Optional
from langchain_core.load import loads, dumps
from langchain_community.chat_models import (
    ChatAnthropic,
    ChatAnyscale,
    ChatBaichuan,
    QianfanChatEndpoint,
    BedrockChat,
    ChatDatabricks,
    ChatDeepInfra,
    ErnieBotChat,
    ChatEverlyAI,
    FakeListChatModel,
    ChatFireworks,
    GigaChat,
    ChatGooglePalm,
    GPTRouter,
    ChatHuggingFace,
    HumanInputChatModel,
    ChatHunyuan,
    ChatJavelinAIGateway,
    JinaChat,
    ChatKonko,
    ChatLiteLLM,
    ChatLiteLLMRouter,
    LlamaEdgeChatService,
    MiniMaxChat,
    ChatMlflow,
    ChatMLflowAIGateway,
    ChatOllama,
    ChatOpenAI,
    AzureChatOpenAI,
    PaiEasChatEndpoint,
    PromptLayerChatOpenAI,
    ChatSparkLLM,
    ChatVertexAI,
    VolcEngineMaasChat,
    ChatYandexGPT,
    ChatZhipuAI,
)
from langchain_community.llms.anthropic import Anthropic
from langchain_community.llms.bedrock import Bedrock
from langchain_community.llms.openai import OpenAI
from langchain_community.llms.openai import AzureOpenAI

# NOTE ON DEPENDENCIES:
# - since Jan 2024, there is https://pypi.org/project/langchain-openai/ which is a separate package and imports openai models.
#   Decided to not make this a dependency of langfuse as few people will have this. Need to match these models manually
# - langchain_community is loaded as a dependency of langchain, so we can use it here


def _extract_model_name(
    serialized: Dict[str, Any],
    **kwargs: Any,
):
    """
    Extracts the model name from the serialized or kwargs object. This is used to get the model names for Langfuse.
    """
    # we have to deal with ChatGoogleGenerativeAI and ChatMistralAI first, as
    # if we run loads(dumps(serialized)) on it, it will throw in case of missing api keys

    model = _extract_model_by_key(
        "ChatGoogleGenerativeAI",
        serialized,
        serialized,
        ["kwargs", "model"],
    )
    if model:
        return model

    model = _extract_model_by_key(
        "ChatMistralAI",
        serialized,
        serialized,
        ["kwargs", "model"],
    )
    if model:
        return model

    # checks if serializations is implemented. Otherwise, this will throw
    if serialized.get("type") != "not_implemented":
        try:
            # try to deserialize the model name from the serialized object
            # https://github.com/langchain-ai/langchain/blob/00a09e1b7117f3bde14a44748510fcccc95f9de5/libs/core/langchain_core/load/load.py#L112

            llm = loads(dumps(serialized))

            # openai models from langchain_openai, separate package, not installed with langchain

            # community models from langchain_community, separate package, installed with langchain
            if isinstance(llm, ChatAnthropic):
                return llm.model

            if isinstance(llm, Anthropic):
                return llm.model

            if isinstance(llm, ChatAnyscale):
                return llm.model_name

            # openai community models
            if isinstance(llm, AzureChatOpenAI):
                return llm.model_name

            if isinstance(llm, ChatOpenAI):
                return llm.model_name

            if isinstance(llm, OpenAI):
                return llm.model_name

            if isinstance(llm, AzureOpenAI):
                return (
                    kwargs.get("invocation_params").get("model")
                    + "-"
                    + llm.serialized["kwargs"]["model_version"]
                )

            if isinstance(llm, ChatBaichuan):
                return llm.model

            if isinstance(llm, QianfanChatEndpoint):
                return llm.model

            if isinstance(llm, BedrockChat):
                return llm.model_id

            if isinstance(llm, Bedrock):
                return llm.model_id

            if isinstance(llm, ChatDatabricks):
                return llm.name

            if isinstance(llm, ChatDeepInfra):
                return llm.model_name

            if isinstance(llm, ErnieBotChat):
                return llm.model_name

            if isinstance(llm, ChatEverlyAI):
                return llm.model_name

            if isinstance(llm, FakeListChatModel):
                return None

            if isinstance(llm, ChatFireworks):
                return llm.model

            if isinstance(llm, GigaChat):
                return llm.model

            if isinstance(llm, ChatGooglePalm):
                return llm.model_name

            if isinstance(llm, GPTRouter):
                # taking the last model from the priority list
                # https://python.langchain.com/docs/integrations/chat/gpt_router

                return (
                    llm.models_priority_list[-1].name
                    if len(llm.models_priority_list) > 0
                    else None
                )

            if isinstance(llm, ChatHuggingFace):
                return llm.model_id

            if isinstance(llm, HumanInputChatModel):
                return llm.name

            if isinstance(llm, ChatHunyuan):
                return llm.name

            if isinstance(llm, ChatJavelinAIGateway):
                return llm.name

            if isinstance(llm, JinaChat):
                return None

            if isinstance(llm, ChatKonko):
                return llm.model

            if isinstance(llm, ChatLiteLLM):
                return llm.model_name

            if isinstance(llm, ChatLiteLLMRouter):
                return llm.model_name

            if isinstance(llm, LlamaEdgeChatService):
                return llm.model

            if isinstance(llm, MiniMaxChat):
                return llm.model

            if isinstance(llm, ChatMlflow):
                return None

            if isinstance(llm, ChatMLflowAIGateway):
                return None

            if isinstance(llm, ChatOllama):
                return llm.model

            if isinstance(llm, PaiEasChatEndpoint):
                return None

            if isinstance(llm, PromptLayerChatOpenAI):
                return None

            if isinstance(llm, ChatSparkLLM):
                return None

            if isinstance(llm, ChatVertexAI):
                return llm.model_name

            if isinstance(llm, VolcEngineMaasChat):
                return llm.model

            if isinstance(llm, ChatYandexGPT):
                return llm.model_name

            if isinstance(llm, ChatZhipuAI):
                return llm.model
        except Exception:
            # using a try .. except block to catch exceptions if the model load above fails as some library is not installed for example
            pass
    # try to extract the model manually

    model = _extract_model_by_key(
        "ChatVertexAI",
        serialized,
        serialized,
        ["kwargs", "model_name"],
    )
    if model:
        return model

    # openai new langchain-openai package
    model = _extract_model_by_key(
        "OpenAI",
        serialized,
        kwargs,
        ["invocation_params", "model_name"],
    )
    if model:
        return model

    model = _extract_model_by_key(
        "ChatOpenAI",
        serialized,
        kwargs,
        ["invocation_params", "model_name"],
    )
    if model:
        return model

    model = _extract_model_by_key(
        "AzureChatOpenAI",
        serialized,
        kwargs,
        ["invocation_params", "model"],
    )
    if model:
        return model

    if serialized.get("id")[-1] == "AzureChatOpenAI":
        if kwargs.get("invocation_params").get("model"):
            return kwargs.get("invocation_params").get("model")

    if serialized.get("id")[-1] == "AzureOpenAI":
        if kwargs.get("invocation_params").get("model_name"):
            return kwargs.get("invocation_params").get("model_name")

        deployment_name = None
        if serialized.get("kwargs").get("openai_api_version"):
            deployment_name = serialized.get("kwargs").get("deployment_version")
        deployment_version = None
        if serialized.get("kwargs").get("deployment_name"):
            deployment_name = serialized.get("kwargs").get("deployment_name")
        return deployment_name + "-" + deployment_version

    # anthropic
    model = _extract_model_by_pattern("Anthropic", serialized, "model", "anthropic")
    if model:
        return model

    # anthropic
    model = _extract_model_by_pattern("ChatAnthropic", serialized, "model")
    if model:
        return model

    # chatongyi
    model = _extract_model_by_pattern("ChatTongyi", serialized, "model_name")
    if model:
        return model

    # Cohere
    model = _extract_model_by_pattern("ChatCohere", serialized, "model")
    if model:
        return model
    model = _extract_model_by_pattern("Cohere", serialized, "model")
    if model:
        return model

    # huggingface
    model = _extract_model_by_pattern("HuggingFaceHub", serialized, "model")
    if model:
        return model

    # anyscale
    model = _extract_model_by_pattern("ChatAnyscale", serialized, "model_name")
    if model:
        return model

    model = _extract_model_by_key(
        "HuggingFacePipeline",
        serialized,
        kwargs,
        ["invocation_params", "model_id"],
    )
    if model:
        return model

    # textgen
    model = _extract_model_by_pattern("TextGen", serialized, "model", "text-gen")
    if model:
        return model

    return None


def _extract_model_with_regex(pattern: str, text: str):
    match = re.search(rf"{pattern}='(.*?)'", text)
    if match:
        return match.group(1)
    return None


def _extract_model_by_pattern(
    id: str, serialized: dict, pattern: str, default: Optional[str] = None
):
    if serialized.get("id")[-1] == id:
        extracted = _extract_model_with_regex(pattern, serialized["repr"])
        return extracted if extracted else default if default else None


def _extract_model_by_key(
    id: str,
    serialized: dict,
    object: dict,
    keys: List[str],
    default: Optional[str] = None,
):
    if serialized.get("id")[-1] == id:
        current_obj = object
        for key in keys:
            current_obj = current_obj.get(key)
            if not current_obj:
                raise ValueError(f"Key {key} not found in {object}")

        return current_obj if current_obj else default if default else None


from dataclasses import dataclass
from typing import Optional, Dict


@dataclass
class AthinaMeta:
    prompt_slug: Optional[str] = None
    context: Optional[Dict] = None
    session_id: Optional[str] = None
    user_query: Optional[str] = None
    environment: Optional[str] = 'production'
    external_reference_id: Optional[str] = None
    customer_id: Optional[str] = None
    customer_user_id: Optional[str] = None
    response_time: Optional[int] = None
    custom_attributes: Optional[Dict] = None
    custom_eval_metrics: Optional[Dict] = None


import functools
import asyncio
import inspect
from typing import Optional, Dict, Any, Callable, TypeVar
from contextvars import ContextVar

from ..trace import Trace

# Context variables to maintain state
current_trace = ContextVar('current_trace', default=None)
current_span = ContextVar('current_span', default=None)

F = TypeVar('F', bound=Callable)

class ObserveDecorator:
    
    def trace(
        self,
        name: Optional[str] = None,
        attributes: Optional[Dict[str, Any]] = None,
        version: Optional[str] = None,
    ):
        def decorator(func: F) -> F:
            # Get the signature of the function
            sig = inspect.signature(func)
            param_names = list(sig.parameters.keys())
            
            @functools.wraps(func)
            async def async_wrapper(*args, **kwargs):
                trace_name = name or func.__name__
                trace = Trace(
                    name=trace_name,
                    attributes=attributes,
                    version=version
                )
                
                # Set trace in context
                token = current_trace.set(trace)
                try:
                    # Map positional args to their parameter names
                    args_dict = {}
                    for i, arg in enumerate(args):
                        if i < len(param_names):
                            args_dict[param_names[i]] = arg
                    
                    # Combine with keyword args
                    input_dict = {**args_dict, **kwargs}
                    
                    # Capture input if available
                    if input_dict:
                        trace._trace.input = input_dict
                    
                    result = await func(*args, **kwargs)
                    
                    # Capture output
                    if result is not None:
                        trace._trace.output = {"result": result}
                    
                    # Set success status if not already set
                    if trace._trace.status is None:
                        trace._trace.status = "success"
                        
                    return result
                except Exception as e:
                    trace._trace.status = "error"
                    if trace._trace.attributes is None:
                        trace._trace.attributes = {}
                    trace._trace.attributes["error"] = str(e)
                    raise
                finally:
                    trace.end()
                    current_trace.reset(token)
            
            @functools.wraps(func)
            def sync_wrapper(*args, **kwargs):
                trace_name = name or func.__name__
                trace = Trace(
                    name=trace_name,
                    attributes=attributes,
                    version=version
                )
                
                # Set trace in context
                token = current_trace.set(trace)
                try:
                    # Map positional args to their parameter names
                    args_dict = {}
                    for i, arg in enumerate(args):
                        if i < len(param_names):
                            args_dict[param_names[i]] = arg
                    
                    # Combine with keyword args
                    input_dict = {**args_dict, **kwargs}
                    
                    # Capture input if available
                    if input_dict:
                        trace._trace.input = input_dict
                    
                    result = func(*args, **kwargs)
                    
                    # Capture output
                    if result is not None:
                        trace._trace.output = {"result": result}
                    
                    # Set success status if not already set
                    if trace._trace.status is None:
                        trace._trace.status = "success"
                        
                    return result
                except Exception as e:
                    trace._trace.status = "error"
                    if trace._trace.attributes is None:
                        trace._trace.attributes = {}
                    trace._trace.attributes["error"] = str(e)
                    raise
                finally:
                    trace.end()
                    current_trace.reset(token)
            
            return async_wrapper if asyncio.iscoroutinefunction(func) else sync_wrapper
        return decorator

    def span(
        self,
        name: Optional[str] = None,
        attributes: Optional[Dict[str, Any]] = None,
        span_type: str = "span",
        version: Optional[str] = None,
    ):
        def decorator(func: F) -> F:
            # Get the signature of the function
            sig = inspect.signature(func)
            param_names = list(sig.parameters.keys())
            
            @functools.wraps(func)
            async def async_wrapper(*args, **kwargs):
                span_name = name or func.__name__
                
                # Get current trace or create new one if none exists
                trace = current_trace.get()
                trace_token = None
                if trace is None:
                    trace = Trace(span_name, attributes=attributes)
                    trace_token = current_trace.set(trace)
                
                # Check if we're inside another span
                parent_span = current_span.get()
                
                # Create and add span
                if parent_span is not None:
                    # If inside another span, create child span
                    span = parent_span.create_span(
                        name=span_name,
                        span_type=span_type,
                        attributes=attributes,
                        version=version
                    )
                else:
                    # Otherwise create a span at the trace level
                    span = trace.create_span(
                        name=span_name,
                        span_type=span_type,
                        attributes=attributes,
                        version=version
                    )
                
                # Set span in context
                span_token = current_span.set(span)
                
                try:
                    # Map positional args to their parameter names
                    args_dict = {}
                    for i, arg in enumerate(args):
                        if i < len(param_names):
                            args_dict[param_names[i]] = arg
                    
                    # Combine with keyword args
                    input_dict = {**args_dict, **kwargs}
                    
                    # Capture input if available
                    if input_dict:
                        span._span.input = input_dict
                    
                    result = await func(*args, **kwargs)
                    
                    # Capture output
                    if result is not None:
                        span._span.output = {"result": result}
                    
                    # Set success status if not already set
                    if span._span.status is None:
                        span._span.status = "success"
                        
                    return result
                except Exception as e:
                    span._span.status = "error"
                    if span._span.attributes is None:
                        span._span.attributes = {}
                    span._span.attributes["error"] = str(e)
                    raise
                finally:
                    span.end()
                    current_span.reset(span_token)
                    if trace_token:
                        # Set success status on trace if not already set
                        if trace._trace.status is None:
                            trace._trace.status = "success"
                        trace.end()
                        current_trace.reset(trace_token)
            
            @functools.wraps(func)
            def sync_wrapper(*args, **kwargs):
                span_name = name or func.__name__
                
                # Get current trace or create new one if none exists
                trace = current_trace.get()
                trace_token = None
                if trace is None:
                    trace = Trace(span_name, attributes=attributes)
                    trace_token = current_trace.set(trace)
                
                # Check if we're inside another span
                parent_span = current_span.get()
                
                # Create and add span
                if parent_span is not None:
                    # If inside another span, create child span
                    span = parent_span.create_span(
                        name=span_name,
                        span_type=span_type,
                        attributes=attributes,
                        version=version
                    )
                else:
                    # Otherwise create a span at the trace level
                    span = trace.create_span(
                        name=span_name,
                        span_type=span_type,
                        attributes=attributes,
                        version=version
                    )
                
                # Set span in context
                span_token = current_span.set(span)
                
                try:
                    # Map positional args to their parameter names
                    args_dict = {}
                    for i, arg in enumerate(args):
                        if i < len(param_names):
                            args_dict[param_names[i]] = arg
                    
                    # Combine with keyword args
                    input_dict = {**args_dict, **kwargs}
                    
                    # Capture input if available
                    if input_dict:
                        span._span.input = input_dict
                    
                    result = func(*args, **kwargs)
                    
                    # Capture output
                    if result is not None:
                        span._span.output = {"result": result}
                    
                    # Set success status if not already set
                    if span._span.status is None:
                        span._span.status = "success"
                        
                    return result
                except Exception as e:
                    span._span.status = "error"
                    if span._span.attributes is None:
                        span._span.attributes = {}
                    span._span.attributes["error"] = str(e)
                    raise
                finally:
                    span.end()
                    current_span.reset(span_token)
                    if trace_token:
                        # Set success status on trace if not already set
                        if trace._trace.status is None:
                            trace._trace.status = "success"
                        trace.end()
                        current_trace.reset(trace_token)
            
            return async_wrapper if asyncio.iscoroutinefunction(func) else sync_wrapper
        return decorator

    def generation(
        self,
        name: Optional[str] = None,
        attributes: Optional[Dict[str, Any]] = None,
        version: Optional[str] = None
    ):
        def decorator(func: F) -> F:
            # Get the signature of the function
            sig = inspect.signature(func)
            param_names = list(sig.parameters.keys())
            
            @functools.wraps(func)
            async def async_wrapper(*args, **kwargs):
                gen_name = name or func.__name__
                
                # Get current trace or create new one if none exists
                trace = current_trace.get()
                trace_token = None
                if trace is None:
                    trace = Trace(gen_name, attributes=attributes)
                    trace_token = current_trace.set(trace)
                
                # Check if we're inside another span
                parent_span = current_span.get()
                
                # Create and add generation
                if parent_span is not None:
                    # If inside another span, create child generation
                    generation = parent_span.create_generation(
                        name=gen_name,
                        attributes=attributes,
                        version=version
                    )
                else:
                    # Otherwise create a generation at the trace level
                    generation = trace.create_generation(
                        name=gen_name,
                        attributes=attributes,
                        version=version
                    )
                
                # Set generation in context
                gen_token = current_span.set(generation)
                
                try:
                    # Map positional args to their parameter names
                    args_dict = {}
                    for i, arg in enumerate(args):
                        if i < len(param_names):
                            args_dict[param_names[i]] = arg
                    
                    # Combine with keyword args
                    input_dict = {**args_dict, **kwargs}
                    
                    # Capture input if available
                    if input_dict:
                        generation._span.input = input_dict
                    
                    result = await func(*args, **kwargs)
                    
                    # Capture output
                    if result is not None:
                        generation._span.output = {"result": result}
                    
                    # Set success status if not already set
                    if generation._span.status is None:
                        generation._span.status = "success"
                        
                    return result
                except Exception as e:
                    generation._span.status = "error"
                    if generation._span.attributes is None:
                        generation._span.attributes = {}
                    generation._span.attributes["error"] = str(e)
                    raise
                finally:
                    generation.end()
                    current_span.reset(gen_token)
                    if trace_token:
                        # Set success status on trace if not already set
                        if trace._trace.status is None:
                            trace._trace.status = "success"
                        trace.end()
                        current_trace.reset(trace_token)
            
            @functools.wraps(func)
            def sync_wrapper(*args, **kwargs):
                gen_name = name or func.__name__
                
                # Get current trace or create new one if none exists
                trace = current_trace.get()
                trace_token = None
                if trace is None:
                    trace = Trace(gen_name, attributes=attributes)
                    trace_token = current_trace.set(trace)
                
                # Check if we're inside another span
                parent_span = current_span.get()
                
                # Create and add generation
                if parent_span is not None:
                    # If inside another span, create child generation
                    generation = parent_span.create_generation(
                        name=gen_name,
                        attributes=attributes,
                        version=version
                    )
                else:
                    # Otherwise create a generation at the trace level
                    generation = trace.create_generation(
                        name=gen_name,
                        attributes=attributes,
                        version=version
                    )
                
                # Set generation in context
                gen_token = current_span.set(generation)
                
                try:
                    # Map positional args to their parameter names
                    args_dict = {}
                    for i, arg in enumerate(args):
                        if i < len(param_names):
                            args_dict[param_names[i]] = arg
                    
                    # Combine with keyword args
                    input_dict = {**args_dict, **kwargs}
                    
                    # Capture input if available
                    if input_dict:
                        generation._span.input = input_dict
                    
                    result = func(*args, **kwargs)
                    
                    # Capture output
                    if result is not None:
                        generation._span.output = {"result": result}
                    
                    # Set success status if not already set
                    if generation._span.status is None:
                        generation._span.status = "success"
                        
                    return result
                except Exception as e:
                    generation._span.status = "error"
                    if generation._span.attributes is None:
                        generation._span.attributes = {}
                    generation._span.attributes["error"] = str(e)
                    raise
                finally:
                    generation.end()
                    current_span.reset(gen_token)
                    if trace_token:
                        # Set success status on trace if not already set
                        if trace._trace.status is None:
                            trace._trace.status = "success"
                        trace.end()
                        current_trace.reset(trace_token)
            
            return async_wrapper if asyncio.iscoroutinefunction(func) else sync_wrapper
        return decorator
    
    def update_current_span(
        self, 
        input: Optional[Any] = None, 
        output: Optional[Any] = None,
        attributes: Optional[Dict[str, Any]] = None,
        status: Optional[str] = None
    ):
        """Update the current span with new information."""
        span = current_span.get()
        if span and span._span:
            if input is not None:
                span._span.input = input
            
            if output is not None:
                span._span.output = output
            
            if attributes is not None:
                if span._span.attributes is None:
                    span._span.attributes = {}
                span._span.attributes.update(attributes)
            
            if status is not None:
                span._span.status = status
    
    def update_current_trace(
        self, 
        input: Optional[Any] = None, 
        output: Optional[Any] = None,
        attributes: Optional[Dict[str, Any]] = None,
        status: Optional[str] = None
    ):
        """Update the current trace with new information."""
        trace = current_trace.get()
        if trace and trace._trace:
            if input is not None:
                trace._trace.input = input
            
            if output is not None:
                trace._trace.output = output
            
            if attributes is not None:
                if trace._trace.attributes is None:
                    trace._trace.attributes = {}
                trace._trace.attributes.update(attributes)
            
            if status is not None:
                trace._trace.status = status

    def get_current_trace(self):
        """Get the current trace object."""
        return current_trace.get()
    
    def get_current_span(self):
        """Get the current span object."""
        return current_span.get()

# Create singleton instance
observe = ObserveDecorator()


import importlib
from dataclasses import dataclass
import datetime
import functools
import traceback
import threading
from typing import Any, Callable, Dict, Optional, List
import time
from .athina_meta import AthinaMeta
from .inference_logger import InferenceLogger
from .api_key import AthinaApiKey
import openai
from .util.token_count_helper import get_prompt_tokens_openai_chat_completion, get_completion_tokens_openai_chat_completion

# Check OpenAI version
openai_version = openai.__version__
version_numbers = tuple(map(int, openai_version.split('.')))


def log_to_athina(result: dict, args: dict, athina_meta: AthinaMeta):
    try:
        prompt_slug = "default"
        context = None
        customer_id = None
        customer_user_id = None
        response_time_ms = None
        session_id = None
        user_query = None
        environment = "production"
        external_reference_id = None
        custom_attributes = None
        custom_eval_metrics = None

        if athina_meta:
            prompt_slug = athina_meta.prompt_slug
            context = athina_meta.context
            response_time_ms = athina_meta.response_time
            customer_id = athina_meta.customer_id
            customer_user_id = athina_meta.customer_user_id
            session_id = athina_meta.session_id
            user_query = athina_meta.user_query
            environment = athina_meta.environment or "production"
            external_reference_id = athina_meta.external_reference_id
            custom_attributes = athina_meta.custom_attributes
            custom_eval_metrics = athina_meta.custom_eval_metrics

        InferenceLogger.log_inference(
            prompt_slug=prompt_slug,
            prompt=args["messages"],
            language_model_id=args["model"],
            response=result,
            context=context,
            response_time=response_time_ms,
            customer_id=customer_id,
            customer_user_id=customer_user_id,
            session_id=session_id,
            user_query=user_query,
            environment=environment,
            external_reference_id=external_reference_id,
            custom_attributes=custom_attributes,
            custom_eval_metrics=custom_eval_metrics,
        )
    except Exception as e:
        print("Exception while logging to Athina: ", e)
        traceback.print_exc()


class OpenAiMiddleware:
    _athina_meta: Optional[AthinaMeta]
    _args: Optional[any]
    _kwargs: Optional[dict]
    athina_response = ''

    def __init__(self):
        pass

    def _with_athina_logging(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # Extract args from OpenAI call
            self._args = args
            self._kwargs = kwargs
            self._athina_meta = kwargs.pop("athina_meta", None)

            # Make the OpenAI call and measure response time
            start_time = time.time()
            openai_response = func(*self._args, **self._kwargs)
            end_time = time.time()
            response_time_ms = int((end_time - start_time) * 1000)

            # Return if no result was returned from OpenAI
            if openai_response is None:
                print("No result was returned from OpenAI")
                return openai_response

            # Construct the Athina Meta object and log to Athina
            try:
                # Construct the Athina Meta object
                if self._athina_meta is not None:
                    self._athina_meta.response_time = response_time_ms
                else:
                    self._athina_meta = AthinaMeta(
                        prompt_slug="default",
                        response_time=response_time_ms,
                        environment="default",
                    )

                return self._response_interceptor(openai_response, ("stream" in self._kwargs and self._kwargs["stream"]))
            except Exception as e:
                print("Exception in Athina logging: ", e)
                traceback.print_exc()
                return openai_response

        return wrapper

    def _response_interceptor(self, response, is_streaming=False,
                            send_response: Callable[[dict], None] = None):
        def generator_intercept_packets():
            for r in response:
                self.collect_stream_inference_by_chunk(r)
                yield r
            self._log_stream_to_athina()

        if is_streaming:
            return generator_intercept_packets()
        else:
            api_thread = threading.Thread(
                target=log_to_athina,
                kwargs={
                    "result": response if version_numbers < (1, 0, 0) else response.model_dump(),
                    "args": self._kwargs,
                    "athina_meta": self._athina_meta,
                },
            )
            api_thread.start()
            return response

    def _get_text_from_stream_chunk(self, stream_chunk):
        """
        gets the text from the stream chunk
        """
        try:
            text = ''
            choices = stream_chunk.get('choices', []) 
            if choices and len(choices) > 0 and 'delta' in choices[0]:
                delta = choices[0].get('delta', {})
                if 'content' in delta and delta['content'] is not None:
                    text = delta.get('content', '')

            return text
        except Exception as e:
            raise e

    def collect_stream_inference_by_chunk(self, stream_chunk):
        """
        collects the inference from the log stream of openai chat completion chunk by chunk
        """
        try:
            if isinstance(stream_chunk, dict):
                self.athina_response += self._get_text_from_stream_chunk(
                    stream_chunk)
            else:
                self.athina_response += self._get_text_from_stream_chunk(
                    stream_chunk.model_dump())
        except Exception as e:
            raise e

    def _log_stream_to_athina(self):
        """
        logs the stream response to the athina
        """
        try:
            prompt_tokens = self._get_prompt_tokens(
                prompt=self._kwargs["messages"], language_model_id=self._kwargs["model"])
            completion_tokens = self._get_completion_tokens(
                response=self.athina_response, language_model_id=self._kwargs["model"])
            if prompt_tokens is not None and completion_tokens is not None:
                total_tokens = prompt_tokens + completion_tokens
            else:
                total_tokens = None 
            payload = {
                'prompt_slug': self._athina_meta.prompt_slug,
                'prompt': self._kwargs["messages"],
                'language_model_id': self._kwargs["model"],
                'response': self.athina_response,
                'response_time': self._athina_meta.response_time,
                'context': self._athina_meta.context,
                'environment': self._athina_meta.environment,
                'customer_id': str(self._athina_meta.customer_id) if self._athina_meta.customer_id is not None else None,
                'customer_user_id': str(self._athina_meta.customer_user_id) if self._athina_meta.customer_user_id is not None else None,
                'session_id': str(self._athina_meta.session_id) if self._athina_meta.session_id is not None else None,
                'user_query': str(self._athina_meta.user_query) if self._athina_meta.user_query is not None else None,
                'external_reference_id': str(self._athina_meta.external_reference_id) if self._athina_meta.external_reference_id is not None else None,
                'prompt_tokens': prompt_tokens,
                'completion_tokens': completion_tokens,
                'total_tokens': total_tokens,
                'custom_attributes': self._athina_meta.custom_attributes,
                'custom_eval_metrics': self._athina_meta.custom_eval_metrics,
            }
            # Remove None fields from the payload
            payload = {k: v for k, v in payload.items() if v is not None}
            InferenceLogger.log_inference(**payload)
        except Exception as e:
            raise e

    def _get_prompt_tokens(self, prompt: List[Dict[str, Any]], language_model_id: str):
        """
        gets the prompt tokens given the prompt for the openai chat model completion
        """
        try:
            tokens = get_prompt_tokens_openai_chat_completion(
                prompt=prompt, language_model_id=language_model_id)
            return tokens
        except Exception as e:
            return None

    def _get_completion_tokens(self, response: str, language_model_id: str):
        """
        gets the completion tokens given the prompt response from the openai chat model completion
        """
        try:
            tokens = get_completion_tokens_openai_chat_completion(
                response=response, language_model_id=language_model_id)
            return tokens
        except Exception as e:
            return None

    # Apply the Athina logging wrapper to OpenAI methods
    def apply_athina(self, openai_instance=None):
        openai_version = openai.__version__
        version_numbers = tuple(map(int, openai_version.split('.')))

        if version_numbers < (1, 0, 0):
            ChatCompletion = importlib.import_module(
                "openai.api_resources").ChatCompletion
            openai_method_name = "create"
            openai_method = getattr(ChatCompletion, openai_method_name)

            # Override the create method with the Athina logging wrapper
            athina_method = self._with_athina_logging(openai_method)
            setattr(ChatCompletion, openai_method_name, athina_method)
        else:
            openai_method_name = "create"
            if openai_instance is not None:
                openai_method = getattr(
                    openai_instance.chat.completions, openai_method_name)

                # Override the chat.completions.create method with the Athina logging wrapper
                athina_method = self._with_athina_logging(openai_method)
                setattr(openai_instance.chat.completions,
                        openai_method_name, athina_method)


middleware = OpenAiMiddleware()
middleware.apply_athina()


if version_numbers > (1, 0, 0):
    # Monkey-patch the constructor of openai.OpenAI
    original_openai_constructor = importlib.import_module(
        "openai").OpenAI.__init__

    def new_openai_constructor(self, *args, **kwargs):
        original_openai_constructor(self, *args, **kwargs)
        middleware.apply_athina(self)

    importlib.import_module(
        "openai").OpenAI.__init__ = new_openai_constructor



import asyncio
import threading
from typing import List, Optional, Dict, Union, Any

from .api_key import AthinaApiKey
from .constants import API_BASE_URL
from .request_helper import RequestHelper


class InferenceLogger(AthinaApiKey):

    @staticmethod
    def log_inference(
            prompt: Optional[Union[List[Dict[str, Any]], Dict[str, Any], str]] = None,
            response: Optional[Any] = None,
            prompt_slug: Optional[str] = None,
            language_model_id: Optional[str] = None,
            environment: Optional[str] = 'production',
            functions: Optional[List[Dict]] = None,
            function_call_response: Optional[Any] = None,
            tools: Optional[Any] = None,
            tool_calls: Optional[Any] = None,
            external_reference_id: Optional[str] = None,
            customer_id: Optional[str] = None,
            customer_user_id: Optional[str] = None,
            session_id: Optional[str] = None,
            user_query: Optional[str] = None,
            prompt_tokens: Optional[int] = None,
            completion_tokens: Optional[int] = None,
            total_tokens: Optional[int] = None,
            response_time: Optional[int] = None,
            context: Optional[Dict] = None,
            expected_response: Optional[str] = None,
            custom_attributes: Optional[Dict] = None,
            custom_eval_metrics: Optional[Dict] = None,
            cost: Optional[float] = None,
            model_options: Optional[dict] = None,
    ) -> None:
        """
            logs prompt run data to athina.

            asynchronously logs structured inference data, safe for performance-critical code.
            errors are suppressed and printed.

            Parameters:
            - prompt (Union[List[Dict[str, Any]], Dict[str, Any], str], optional): The main prompt or message data, supporting string or structured list of messages.
            - response (Any, optional): The response to the prompt, can be of any type.
            - prompt_slug (str, optional): Unique slug identifying the prompt type.
            - language_model_id (str, optional): Identifier for the language model used.
            - environment (str, optional): Operational environment, e.g., 'production' or 'development'.
            - functions (List[Dict[str, Any]], optional): Functions metadata used in processing, as an array of function schemas.
            - function_call_response (Any, optional): Response object or data resulting from function calls.
            - tools (Any, optional): Metadata for tools used in inference, if any.
            - tool_calls (Any, optional): Metadata for tool call results.
            - external_reference_id (str, optional): External reference for tracking, nullable.
            - customer_id (str, optional): Identifier for the customer using the service, nullable.
            - customer_user_id (str, optional): Unique user identifier for the customer’s user, nullable.
            - session_id (str, optional): ID for the current session, nullable.
            - user_query (str, optional): Original query from the user, nullable.
            - prompt_tokens (int, optional): Number of tokens in the prompt input.
            - completion_tokens (int, optional): Number of tokens in the model’s output completion.
            - total_tokens (int, optional): Sum of tokens used (prompt + completion).
            - response_time (int, optional): Time taken to generate a response, in milliseconds.
            - context (Union[Dict[str, Any], str], optional): Additional context data as a dictionary or string, nullable.
            - expected_response (str, optional): The expected response for comparison or validation, nullable.
            - custom_attributes (Dict[str, Any], optional): Additional attributes for custom data, structured as key-value pairs.
            - custom_eval_metrics (Dict[str, Any], optional): Custom evaluation metrics for assessing response quality.
            - cost (float, optional): Inference cost, if calculated.
            - model_options (Dict[str, Any], optional): Configuration options for the model run, including:
              - `temperature` (float, optional): Sampling temperature.
              - `max_completion_tokens` (int, optional): Maximum tokens allowed in response.
              - `stop` (Union[str, List[str]], optional): Stop sequence(s) to halt generation.
              - `top_p` (float, optional): Top-p sampling parameter.
              - `extra_options` (Dict[str, Any], optional): Any additional options for model customization.

            Returns:
            - None: The method does not return any value.

            Raises:
            - None: errors are suppressed and printed.
            """
        try:
            args = (
                prompt, response, prompt_slug, language_model_id, environment, functions, function_call_response, tools,
                tool_calls, external_reference_id, customer_id, customer_user_id, session_id, user_query, prompt_tokens,
                completion_tokens, total_tokens, response_time, context, expected_response, custom_attributes, cost,
                custom_eval_metrics, model_options)
            threading.Thread(target=lambda: asyncio.run(InferenceLogger._log_inference_asynchronously(*args))).start()
        except Exception as e:
            print("Error in logging inference to Athina: ", str(e))

    @staticmethod
    async def _log_inference_asynchronously(
            prompt, response, prompt_slug, language_model_id, environment, functions, function_call_response, tools,
            tool_calls, external_reference_id, customer_id, customer_user_id, session_id, user_query, prompt_tokens,
            completion_tokens, total_tokens, response_time, context, expected_response, custom_attributes, cost,
            custom_eval_metrics, model_options
    ) -> None:
        """
        logs the llm inference to athina
        """
        try:
            payload = {
                'prompt': prompt,
                'response': response,
                'prompt_slug': prompt_slug,
                'language_model_id': language_model_id,
                'functions': functions,
                'function_call_response': function_call_response,
                'tools': tools,
                'tool_calls': tool_calls,
                'response_time': response_time,
                'context': context,
                'environment': environment,
                'customer_id': str(customer_id) if customer_id is not None else None,
                'customer_user_id': str(customer_user_id) if customer_user_id is not None else None,
                'session_id': str(session_id) if session_id is not None else None,
                'user_query': str(user_query) if user_query is not None else None,
                'external_reference_id': str(external_reference_id) if external_reference_id is not None else None,
                'prompt_tokens': prompt_tokens,
                'completion_tokens': completion_tokens,
                'total_tokens': total_tokens,
                'expected_response': expected_response,
                'custom_attributes': custom_attributes,
                'custom_eval_metrics': custom_eval_metrics,
                'cost': cost,
                'model_options': model_options,
            }
            # Remove None fields from the payload
            payload = {k: v for k, v in payload.items() if v is not None}
            RequestHelper.make_post_request(endpoint=f'{API_BASE_URL}/api/v1/log/inference', payload=payload, headers={
                'athina-api-key': InferenceLogger.get_api_key(),
            })
        except Exception as e:
            print("Error in logging inference to Athina: ", str(e))



from langchain.globals import get_debug
from athina_logger.api_key import AthinaApiKey
 
from typing import Any, Dict, List, Optional, Sequence, Union
from uuid import UUID, uuid4
from langchain_core.callbacks import BaseCallbackHandler
from langchain_core.messages.ai import AIMessageChunk
from langchain_core.prompt_values import ChatPromptValue
from langchain.schema.agent import AgentAction, AgentActionMessageLog, AgentFinish
from langchain.schema.document import Document
from langchain_core.outputs import (
    ChatGeneration,
    LLMResult,
)
from langchain_core.messages import (
    AIMessage,
    BaseMessage,
    ChatMessage,
    HumanMessage,
    SystemMessage,
    ToolMessage,
    FunctionMessage,
)
from athina_logger.tracing.trace import Trace
from athina_logger.util.extract_model import _extract_model_name

class LangchainCallbackHandler(
    BaseCallbackHandler, AthinaApiKey
):
    def __init__(
        self,
        trace_name: Optional[str] = None,
        version: Optional[str] = None,
    ) -> None: 
        _debug("LangchainCallbackHandler.__init__")
        self.version = version
        self.trace_name = trace_name
        self.trace_run_id = None
        self.runs = {}
        self.trace: Trace = None 

    def on_llm_new_token(
        self,
        token: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        """Run on new LLM token. Only available when streaming is enabled."""
        _debug(
            f"on llm new token: run_id: {run_id} parent_run_id: {parent_run_id}"
        )

    def _get_athina_run_name(self, serialized: Dict[str, Any], **kwargs: Any) -> str:
        """
        Retrieves the 'run_name' for an entity prioritizing the 'name' key in 'kwargs' or falling
        back to the 'name' or 'id' in 'serialized'. Defaults to "<unknown>" if none are available.

        Args:
            serialized (Dict[str, Any]): A dictionary containing the entity's serialized data.
            **kwargs (Any): Additional keyword arguments, potentially including the 'name' override.

        Returns:
            str: The determined athina run name for the entity.
        """
        # Check if 'name' is in kwargs and not None, otherwise use default fallback logic
        if "name" in kwargs and kwargs["name"] is not None:
            return str(kwargs["name"])

        # Fallback to serialized 'name', 'id', or "<unknown>"
        return serialized.get("name", serialized.get("id", ["<unknown>"])[-1])

    def on_chain_start(
        self,
        serialized: Dict[str, Any],
        inputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any) -> Any:
        _debug(
            f"on chain start: run_id: {run_id} parent_run_id: {parent_run_id}, name {serialized.get('name', serialized.get('id', ['<unknown>'])[-1])}"
        )
        self._generate_trace(
            serialized=serialized,
            inputs=inputs,
            run_id=run_id,
            parent_run_id=parent_run_id,
            tags=tags,
            metadata=metadata,
            version=self.version,
            **kwargs,
        )
        name = self._get_athina_run_name(serialized, **kwargs) 
        if "intermediate_steps" in inputs:
            inputs["intermediate_steps"] = [
                { 
                    "tool_input":agent_action_message.tool_input,
                    "log":agent_action_message.log
                } for agent_action_message in inputs["intermediate_steps"]
            ]
        if parent_run_id is None:
            self.runs[run_id] = self.trace.create_span(name=name, input=inputs, version=self.version)

        if parent_run_id is not None:
            self.runs[run_id] = self.runs[parent_run_id].create_span(name=name, input=inputs, version=self.version) 

    def convert_outputs_to_dict(self, outputs: Dict[str, Any]) -> Dict:
        if "intermediate_steps" in outputs:
            outputs["intermediate_steps"] = [
                { 
                    "tool_input":agent_action_message.tool_input,
                    "log":agent_action_message.log
                } for agent_action_message in outputs["intermediate_steps"]
            ]
        if type(outputs) == AgentActionMessageLog:
            outputs = { 
                "tool_input": outputs.tool_input,
                "log": outputs.log
            }
        if type(outputs) == ChatPromptValue:
            outputs = { 
                "messages": outputs.to_string(),
            }
        if type(outputs) == AgentFinish:
            outputs = { 
                "return_values": outputs.return_values,
            }
        if type(outputs) == FunctionMessage:
            outputs = {
                "content": outputs.content
            }
        if type(outputs) == list and all(isinstance(item, AIMessageChunk) for item in outputs):
            outputs = [
                { 
                    "type":ai_message_chunk.type
                } for ai_message_chunk in outputs
            ]
        if "agent_scratchpad" in outputs:
            outputs["agent_scratchpad"] = [
                { 
                    "type":ai_message_chunk.type
                } for ai_message_chunk in outputs["agent_scratchpad"]
            ]
        if outputs is None or type(outputs) is not dict:
            outputs = {}
        return outputs

    def on_chain_end(
        self,
        outputs: Dict[str, Any],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        try:
            _debug(
                f"on chain end: run_id: {run_id} parent_run_id: {parent_run_id}"
            )
            if run_id not in self.runs:
                raise Exception("run not found")
            outputs = self.convert_outputs_to_dict(outputs)
            self._update_run(run_id, outputs, None)
            self.runs[run_id].end()
            if self.trace_run_id == run_id:
                self.trace.end()
        except Exception as e:
            _debug(e)

    def on_chain_error(
        self,
        error: Union[Exception, KeyboardInterrupt],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        try:
            _debug(
                f"on chain error: run_id: {run_id} parent_run_id: {parent_run_id}"
            )
            self._update_run(run_id, None, error)
            self.runs[run_id].end()

        except Exception as e:
            _debug(e)

    def on_agent_action(
        self,
        action: AgentAction,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        """Run on agent action."""
        try:
            _debug(
                f"on agent action: run_id: {run_id} parent_run_id: {parent_run_id}"
            )

            if run_id not in self.runs:
                raise Exception("run not found")

            self._update_run(run_id, {"agent_action" : action.tool_input}, None)
            self.runs[run_id].end()

        except Exception as e:
            _debug(e)

    def on_agent_finish(
        self,
        finish: AgentFinish,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        try:
            _debug(
                f"on agent finish: run_id: {run_id} parent_run_id: {parent_run_id}"
            )
            if run_id not in self.runs:
                raise Exception("run not found")

            self._update_run(run_id, {"agent_finish": finish.return_values}, None)
            self.runs[run_id].end()

        except Exception as e:
            _debug(e)

    def on_chat_model_start(
        self,
        serialized: Dict[str, Any],
        messages: List[List[BaseMessage]],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        try:
            _debug(
                f"on chat model start: run_id: {run_id} parent_run_id: {parent_run_id}"
            )
            if metadata is None:
                metadata = {}
            metadata['is_chat_model'] = True
            self._on_llm_action(
                serialized,
                run_id,
                _flatten_comprehension(
                    [self._create_message_dicts(m) for m in messages]
                ),
                parent_run_id,
                tags=tags,
                metadata=metadata,
                **kwargs,
            )
        except Exception as e:
            _debug(e)

    def on_retriever_start(
        self,
        serialized: Dict[str, Any],
        query: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        try:
            _debug(
                f"on retriever start: run_id: {run_id} parent_run_id: {parent_run_id}"
            )

            if parent_run_id is None or parent_run_id not in self.runs:
                raise Exception("parent run not found")

            self.runs[run_id] = self.runs[parent_run_id].create_span(
                name=self._get_athina_run_name(serialized, **kwargs),
                input={"query":query},
                attributes=self._join_tags_and_metadata(tags, metadata),
                version=self.version,
            )

        except Exception as e:
            _debug(e)

    def on_retriever_end(
        self,
        documents: Sequence[Document],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        try:
            _debug(
                f"on retriever end: run_id: {run_id} parent_run_id: {parent_run_id}"
            )
            if run_id is None or run_id not in self.runs:
                raise Exception("run not found")
            documents_json = {"documents" : [doc.page_content for doc in documents]}
            self._update_run(run_id, documents_json, None)
            self.runs[run_id].end()
        except Exception as e:
            _debug(e)

    def on_retriever_error(
        self,
        error: Union[Exception, KeyboardInterrupt],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        """Run when Retriever errors."""
        try:
            _debug(
                f"on retriever error: run_id: {run_id} parent_run_id: {parent_run_id}"
            )

            if run_id is None or run_id not in self.runs:
                raise Exception("run not found")
            self._update_run(run_id, None, error)
            self.runs[run_id].end()

        except Exception as e:
            _debug(e)

    def on_tool_start(
        self,
        serialized: Dict[str, Any],
        input_str: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        try:
            _debug(
                f"on tool start: run_id: {run_id} parent_run_id: {parent_run_id}"
            )
            if parent_run_id is None or parent_run_id not in self.runs:
                raise Exception("parent run not found")
            self.runs[run_id] = self.runs[parent_run_id].create_span(
                name=self._get_athina_run_name(serialized, **kwargs),
                input={"input_str": input_str},
                attributes=self._join_tags_and_metadata(tags, metadata),
                version=self.version,
            )
        except Exception as e:
            _debug(e)

    def on_tool_end(
        self,
        output: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        try:
            _debug(
                f"on tool end: run_id: {run_id} parent_run_id: {parent_run_id}"
            )
            if run_id is None or run_id not in self.runs:
                raise Exception("run not found")
            self._update_run(run_id, {"tool_output" : output}, None)
            self.runs[run_id].end()

        except Exception as e:
            _debug(e)

    def on_tool_error(
        self,
        error: Union[Exception, KeyboardInterrupt],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        try:
            _debug(
                f"on tool error: run_id: {run_id} parent_run_id: {parent_run_id}"
            )
            if run_id is None or run_id not in self.runs:
                raise Exception("run not found")
            self._update_run(run_id, None, error)
            self.runs[run_id].end()

        except Exception as e:
            _debug(e)

    def on_llm_start(
        self,
        serialized: Dict[str, Any],
        prompts: List[str],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        try:
            _debug(
                f"on llm start: run_id: {run_id} parent_run_id: {parent_run_id}"
            )
            _debug("metadata:")
            _debug(metadata)
            self._on_llm_action(
                serialized,
                run_id,
                prompts[0] if len(prompts) == 1 else prompts,
                parent_run_id,
                tags=tags,
                metadata=metadata,
                **kwargs,
            )
        except Exception as e:
            _debug(e)

    def on_llm_end(
        self,
        response: LLMResult,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        try:
            _debug(
                f"on llm end: run_id: {run_id} parent_run_id: {parent_run_id} response: {response} kwargs: {kwargs}"
            )
            if run_id not in self.runs:
                raise Exception("Run not found, something went wrong.")
            else:
                generation = response.generations[-1][-1]
                extracted_response = (
                    self._convert_message_to_dict(generation.message)
                    if isinstance(generation, ChatGeneration)
                    else _extract_raw_response(generation)
                )
                if isinstance(extracted_response, Dict):
                    extracted_response = extracted_response.get('content', None)

                llm_usage = (
                    None
                    if response.llm_output is None
                    or not response.llm_output["token_usage"]
                    else response.llm_output["token_usage"]
                )
                self.runs[run_id].update(
                    prompt_tokens=llm_usage["prompt_tokens"] if llm_usage else None, 
                    completion_tokens=llm_usage["completion_tokens"] if llm_usage else None,
                    total_tokens=llm_usage["total_tokens"] if llm_usage else None, 
                    response=extracted_response
                )
                self.runs[run_id].end()
        except Exception as e:
            import traceback
            traceback.print_exc()
            _debug(e)

    def on_llm_error(
        self,
        error: Union[Exception, KeyboardInterrupt],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        try:
            _debug(
                f"on llm error: run_id: {run_id} parent_run_id: {parent_run_id}"
            )
            self._update_run(run_id, None, error)
            self.runs[run_id].end()

        except Exception as e:
            _debug(e)

    def _on_llm_action(
        self,
        serialized: Dict[str, Any],
        run_id: UUID,
        prompts: List[Any],
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ):
        try:
            self._generate_trace(
                serialized,
                inputs=prompts[0] if len(prompts) == 1 else prompts,
                run_id=run_id,
                parent_run_id=parent_run_id,
                tags=tags,
                metadata=metadata,
                version=self.version,
                kwargs=kwargs,
            ) 
            name = self._get_athina_run_name(serialized, **kwargs)
            attributes = { 
                'is_chat_model' : metadata.get('is_chat_model', False),
                'prompt_slug': metadata.get('prompt_slug', None),
                'user_query': metadata.get('user_query', None),
                'context': metadata.get('global_context', None),
                'prompt': prompts,
                'session_id': metadata.get('session_id', None),
                'customer_id': metadata.get('customer_id', None),
                'customer_user_id': metadata.get('customer_user_id', None),
                'external_reference_id': metadata.get('external_reference_id', None),
                'custom_attributes': metadata.get('custom_attributes', None),
                'language_model_id': _extract_model_name(serialized, **kwargs)
            }
            prompt_slug = metadata.get('prompt_slug', None)
            _debug(f"Creating generation with name: {name} and attributes: {attributes}")
            if parent_run_id in self.runs:
                self.runs[run_id] = self.runs[parent_run_id].create_generation(name=name, attributes=attributes, version=self.version, prompt_slug=prompt_slug)
            else:
                self.runs[run_id] = self.trace.create_generation(name=name, attributes=attributes, version=self.version, prompt_slug=prompt_slug)

        except Exception as e:
            _debug(e)


    def _generate_trace(
        self,
        serialized: Dict[str, Any],
        inputs: Union[Dict[str, Any], List[str], str, None],
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ):
        try:
            class_name = self._get_athina_run_name(serialized, **kwargs)
            # Initialise trace by creating a trace if it does not exist
            if self.trace is None:
                trace = Trace(
                    name=self.trace_name if self.trace_name is not None else class_name,
                    attributes=metadata,
                    version=self.version,
                )
                self.trace = trace 
                self.trace_run_id = run_id
        except Exception as e:
            _debug(e)

    def _join_tags_and_metadata(
        self,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> Optional[Dict[str, Any]]:
        if tags is None and metadata is None:
            return None
        elif tags is not None and len(tags) > 0:
            final_dict = {"tags": tags}
            if metadata is not None:
                final_dict.update(metadata)  # Merge metadata into final_dict
            return final_dict
        else:
            return metadata
 
    def _update_run(self, run_id: str, output: Dict, error: Optional[Exception] = None):
        """Update the trace/span with the output of the current run."""
        if self.trace is not None and self.runs[run_id] is not None:
            if error is not None:
                self.runs[run_id].update(status="ERROR",attributes={"status_message": str(error)})
            else:
                self.runs[run_id].update(output=output)

    def _convert_message_to_dict(self, message: BaseMessage) -> Dict[str, Any]:
        # assistant message
        if isinstance(message, HumanMessage):
            message_dict = {"role": "user", "content": message.content}
        elif isinstance(message, AIMessage):
            message_dict = {"role": "assistant", "content": message.content}
        elif isinstance(message, SystemMessage):
            message_dict = {"role": "system", "content": message.content}
        elif isinstance(message, ToolMessage):
            message_dict = {"role": "tool", "content": message.content}
        elif isinstance(message, FunctionMessage):
            message_dict = {"role": "function", "content": message.content}
        elif isinstance(message, ChatMessage):
            message_dict = {"role": message.role, "content": message.content}
        else:
            raise ValueError(f"Got unknown type {message}")
        if "name" in message.additional_kwargs:
            message_dict["name"] = message.additional_kwargs["name"]

        if message.additional_kwargs:
            message_dict["additional_kwargs"] = message.additional_kwargs

        return message_dict

    def _create_message_dicts(
        self, messages: List[BaseMessage]
    ) -> List[Dict[str, Any]]:
        return [self._convert_message_to_dict(m) for m in messages]


def _extract_raw_response(last_response):
    """Extract the response from the last response of the LLM call."""
    # We return the text of the response if not empty, otherwise the additional_kwargs
    # Additional kwargs contains the response in case of tool usage
    return (
        last_response.text.strip()
        if last_response.text is not None and last_response.text.strip() != ""
        else last_response.message.additional_kwargs
    )

def _flatten_comprehension(matrix):
    return [item for row in matrix for item in row]

def _debug(msg):
    if get_debug():
        print(msg)